// define GameError as subclass of Error
function GameError(message) {
    
    this.name = "GameError"
    this.message = message;
}
GameError.prototype = Object.create(Error.prototype)
GameError.prototype.constructor = GameError

Game = function(args) {
    
	// process arguments passed to the constructor
    var display_grid = args.display_grid,
    	display_expression = args.display_expression,
    	display_stats = args.display_stats,
    	display_dialog = args.display_dialog,
        defaultLanguage = args.default_lang !== undefined ? args.default_lang : "us-EN",
    	size = args.size !== undefined ? args.size : 4,
    	maximal_value = args.maximal_value !== undefined ? args.maximal_value : 10,
    	speechEnabled = args.speech_enabled !== undefined ? args.speech_enabled : true;
    
    // locate DOM elements used in the game 
    var $grid = $(display_grid),
        $expression = $(display_expression),
        $stats = $(display_stats),
        $dialog = $(display_dialog);
    
    // variables needed for game play
    var number,
    	numbers,
        expression,
        stats = {},
        operators = ['+', '-', '*', '/'];
    
    var numberClickEnabled = true,
        expressionClickEnabled = true;
    
    // initialize success messages
    var messages = [
        "Goed gedaan !!",
        "Applaus",
        "Een dikke pluim",
        "Zoentjes !!",
        "Dubbele dikke duim !!",
        "Hoera",
        "Proficiat !!",
        "Koel",
        "Sjamajee !!"
    ]
    
    // add click handler to expression
    if (speechSupported && speechEnabled) {
        $expression.click(function() {
        	speakExpression();
        });
    }
    
    restart();    
    
    return {
        restart: restart,
        addOperator: addOperator,
        removeOperator: removeOperator,
        setMaximalValue: setMaximalValue,
        enableSpeech: enableSpeech
    };
        
    function restart() {
    	
    	// generate a new grid of numbers
    	newGrid();
    	
        // display the grid
        displayGrid();
           
        // generate a new expression
        newExpression();
        
        // display the expression
        displayExpression();
        
        // clear stats
    	clearStats();
    	
        // display the stats
        displayStats();

    }
    
    function newNumber(pos, different) {
    	
    	var minimal_value = 0,
    		number;
    	
    	numbers[pos] = undefined;
    	
    	if (size * size > (maximal_value - minimal_value + 1)) {
    		different = false;
    	}
    	
    	number = minimal_value + randomNumber(maximal_value - minimal_value);
    	while (different && numbers.indexOf(number) !== -1) {
    		number = minimal_value + randomNumber(maximal_value - minimal_value);
    	}
    	
    	numbers[pos] = number;
    	
    }
    
    function newGrid() {
    	
    	var minimal_value = 0;
    	
    	numbers = randomNumbers(
    		size * size, 
    		minimal_value, 
    		maximal_value, 
    		(maximal_value - minimal_value) < size * size ? false : true
    	);
    	
    }
    
    function displayGrid() {
    	
    	var r, c, pos;
        
        // generate HTML code for the grid
        var grid = "<table>";
        var cell = '<td class="game-number" data-pos="{pos}">{number}</td>';
        for (var r = 0; r < size; r += 1) {
            grid += "<tr>";
            for (var c = 0; c < size; c += 1) {
            	pos = r * size + c;
                grid += cell.format({pos: pos, number: numbers[pos]});            
            }
            grid += "</tr>";
        }
        grid += "</table>";
        
        // display grid
        $grid.html(grid);
        
        // add click handlers to letters of game grid
        $(".game-number").click(clickNumber);
        
    }
        
    function updateGrid() {
    	
    	var pos;

    	for (pos = 0; pos < numbers.length; pos += 1) {
    		if (numbers[pos] > maximal_value) {
    			newNumber(pos, true);
    		}
    	}

    	displayGrid();
    	
    }
    
    function newExpression() {
    	
		assert(
			operators.length > 0, 
			'er moet minstens één bewerking geselecteerd worden'
		);

    	var operator = randomElement(operators);
    	
        // generate a random expression for a random number on the grid
    	number = randomElement(numbers);
    	
    	expression = randomExpression(operator, number, maximal_value);
    	expression.push(operator);
        	
    }
    
    function displayExpression() {
    	
		assert(
			operators.length > 0, 
			'er moet minstens één bewerking geselecteerd worden'
		);

    	exp = expression.slice();
    	exp[2] = {
    			'+': '&plus;',
    			'-': '&minus;',
    			'*': '&times;',
    			'/': '&divide;',
        	}[exp[2]]
            
        $expression.html(exp[0] + ' ' + exp[2] + ' ' + exp[1]);

    }
    
    function updateExpression() {
    	
		assert(
			operators.length > 0, 
			'er moet minstens één bewerking geselecteerd worden'
		);

    	if (
    			number === undefined ||
    			number > maximal_value ||
        		(
        			expression !== undefined && (
	            		expression[0] > maximal_value ||
	            		expression[1] > maximal_value ||
	            		operators.indexOf(expression[2]) === -1
        			)
        		)
        ) {
    		newExpression();
    	}
		displayExpression();
            	
    }
    
    function clearStats() {
    	
    	var operator, 
    	    operators = ['+', '-', '*', '/'];
    	
    	stats = {};
    	for (operator of operators) {
    		stats[operator] = [0, 0];
    	}
    	
    }
    
    function displayStats() {
        
        // generate HTML code for the list of words
        var table = '<table class="table table-striped"><tbody>',
            operators = ['+', '-', '*', '/'],
            operator_names = ['optelling', 'aftrekking', 'vermenigvuldiging', 'deling']
        	correct = 0,
        	wrong = 0;
        var row = '<tr class="game-statistic {footer}"><td class="operator"><span>{name}</span></td><td class="correct">{correct}</td><td class="wrong">{wrong}</td></tr>';
        table += operators.map(function(operator, index) {
        	correct += stats[operator][0];
        	wrong += stats[operator][1];
            return row.format({
                name: operator_names[index], 
                correct: stats[operator][0],
                wrong: stats[operator][1],
                footer: ''
            });
        }).join("");
        table += row.format({
            name: 'totaal', 
            correct: correct,
            wrong: wrong,
            footer: 'info'
        });
        table += "</tbody></table>";
        
        // display words
        $stats.html(table);
        
    }
    
    function addOperator(operator) {
    	
    	if (operators.indexOf(operator) === -1) {
    		operators.push(operator);
    	}
    	
    	if (operators.length === 1) {
    		    	
        	// update the expression if it is inconsistent with the new settings
        	updateExpression();
        	
    	}
    	
    }
    
    function removeOperator(operator) {
    	
    	var index = operators.indexOf(operator);
    	
    	if (index !== -1) {
    		operators.splice(index, 1);
    	}
    	
    	try {
    		    		
    		assert(
    			operators.length > 0, 
    			'er moet minstens één operator geactiveerd worden'
    		);
    		
        	// update the expression if it is inconsistent with the new settings
        	updateExpression();
        	
    	} catch(e) {
    		
    		$expression.html('');
    		
    	}
    	
    }
    
    function setMaximalValue(value) {
    	
    	// set the new maximal values
    	maximal_value = value;
    	
    	// update the values in the grid that are inconsistent with the new
    	// settings
    	updateGrid();
    	
    	// update the epxression if it is inconsistent with the new settings
    	updateExpression();
    }
    
    function enableSpeech(enabled) {
    	speechEnabled = enabled;
    }
    
    function textExpression() {
    	
    	exp = expression.slice();
    	exp[2] = {
    			'+': '+',
    			'-': 'minus',
    			'*': 'maal',
    			'/': 'gedeeld door',
        	}[exp[2]]

    	return exp[0] + ' ' + exp[2] + ' ' + exp[1];
    	
    }
    
    function speakExpression() {
    	
    	// do nothing while already speaking
    	if (!expressionClickEnabled || !speechEnabled) { return; }
    	
    	// disable speak while speaking
    	expressionClickEnabled = false;
        
        text2speech(
        	textExpression(), 
        	defaultLanguage,
        	0.75,
        	function() {
        		expressionClickEnabled = true;
        	}
        );
        
    }
    
    function clickNumber() {
    	
    	// do nothing if number clicks are disabled
    	if (!numberClickEnabled) { return; }
    	
        // disable number clicks
    	numberClickEnabled = false;
        
        // find cell that has been clicked
        var $cell = $(this),
        	correct,
        	$correct,
        	$wrong,
            answer = +$cell.text(),
            operator = expression[2],
            fadeDuration = 250,
            defaultDelay = 1000,
            message;
        
        if (answer === number) {
        	
        	// set the correct and wrong cells
        	$correct = $cell;
        	correct = +$cell.data('pos');
        	
        	// update stats for the operator
        	stats[operator][0] += 1;
        	
        	// set message for spoken result
        	message = 'Juist... {expression} is {result}.'.format({
        		expression: textExpression(),
        		result: number
        	});
        	        	
        } else {
        	
        	// set the correct and wrong cells
        	correct = numbers.indexOf(number);
        	$correct = $('.game-number[data-pos="{pos}"]'.format({pos: correct}));
        	$wrong = $cell;
        	
        	// update stats for the operator
        	stats[operator][1] += 1;
        	
        	// set message for spoken result
        	message = 'Fout... {expression} is {result}.'.format({
        		expression: textExpression(),
        		result: number
        	});
        	        	
        }
        
        displayStats();
        
        function fadeIn($cell, color, duration, callback) {
        	
        	// fade in to target color
        	$cell.animate(
        		{backgroundColor: color}, 
        		{
        			duration: duration,
        			// call callback function
        			complete: callback
        		}
        	);        	
        }
        
        function fadeOut($cell, color, duration, callback) {
        	
        	// fade in to target color
        	$cell.animate(
        		{backgroundColor: color}, 
        		{
        			duration: duration,
        			// call callback function
        			complete: function() {

        				// remove inline style of background color to
						// allow the background color to be set again
						// using external styles
        				$cell.css("backgroundColor", "");
        				
            			// call callback function
        				if (callback !== undefined) {        					
            				callback.apply(this, this.arguments);
        				}
        				
        			}
        		}
        	);        	
        }
        
        // update grid and generate new expression
        function nextExpression() {
        	
            // generate new number at the correct position
            newNumber(correct, true);
            
        	// generate new grid
            displayGrid();
            
            // generate and display a new expression
            newExpression();
            displayExpression();

            // enable number clicks
        	numberClickEnabled = true;

    		// fade in cells
        	var backgroundColor = $grid.css('backgroundColor');
        	if ($wrong) { fadeOut($wrong, backgroundColor, fadeDuration); }
        	fadeOut($correct, backgroundColor, fadeDuration);
        	
        }
        
		// fade in cells
    	if ($wrong) { fadeIn($wrong, '#FF3300', fadeDuration); }
    	fadeIn($correct, '#33CC33', fadeDuration);

        // spoken result
        if (speechSupported && speechEnabled) {
            text2speech(
            	message, 
            	defaultLanguage,
            	0.75,
            	nextExpression
            );        	        	
        } else {
            setTimeout(nextExpression, defaultDelay);        	
        }
        
    }
    
    // show modal dialog
    function dialog(type, message, title) {
        
        // dictionary that maps different types of dialogs onto default titles
        // for each dialog type
        var titles = {
            "primary": "Informatie",
            "success": "Succes",
            "info": "Informatie",
            "warning": "Waarschuwing",
            "danger": "Gevaar",
        }
        
        // select element that represents dialog
        var $dialog_header = $dialog.find('.modal-header');
            $dialog_body = $dialog.find('.modal-body');
            $dialog_btn = $dialog.find('.btn');
        
        // remove previous color
        Object.keys(titles).forEach(function (type) {
            $dialog_header.removeClass('bg-' + type);
            $dialog_body.removeClass('text-' + type);
            $dialog_btn.removeClass('btn-' + type);
        });
        
        // set new color
        if (titles.hasOwnProperty(type)) {
            $dialog_header.addClass('bg-' + type);            
            $dialog_body.addClass('text-' + type);            
            $dialog_btn.addClass('btn-' + type);
        }
        
        // set message
        $dialog_body.find('p').html(message);
        
        // set title
        title = title || titles[type] || "";
        $dialog_header.find('h4').html(title);            
        
        // display dialog
        $dialog.modal(); 
        
    }     
    
    // check if a given condition holds, and show an alert dialog if this is not
    // the case
    function assert(condition, message) {
    	
        if (!condition) {
            
            // dialoogvenster weergeven
            dialog("warning", message);                            

            throw {
                name: 'GameConfigurationError',
                message: message
            };            
        }
        
    }
    
}

// check if speech is supported by the browser
if ('speechSynthesis' in window) {

	speechSupported = true;
	
	function loadVoices() {
		
		// predefine the speaker as a global variable
	    speaker = new SpeechSynthesisUtterance();

		// pick the first Dutch voice
	    var voices = window.speechSynthesis.getVoices();
	    for (voice of voices) {
	    	if (voice.name.toLowerCase().indexOf('dutch') !== -1 || voice.name.toLowerCase().indexOf('nederlands') !== -1) {
	    		speaker.voice = voice;
	    		break;
	    	}
	    }
	    
	}
	
	// load voices synchronously
	loadVoices();
	
	// load voices asynchronously (Chrome)
	window.speechSynthesis.onvoiceschanged = function() {
		loadVoices();		
	}
	
    
} else {
	
	speechSupported = false;
	
}

function text2speech(message, lang, rate, callback) {

	if (speechSupported) {
		
	    try {
	    	
	        // set pronounciation parameters
	        speaker.text = message;
	        speaker.lang = lang || 'en-US';        
	        speaker.rate = rate || 0.75;

	        // set callback to be executed after the speech
	        speaker.onend = function(e) {
	            if (callback) { 
	            	callback(); 
	            }
	        };

	        // set callback to be executed after the speech
	        speaker.onerror = function(e) {
	            if (callback) { 
	            	callback(); 
	            }
	        };

	        // text to speech
	        window.speechSynthesis.speak(speaker);    

	    } catch(e) {
	    	
	    	speechSupported = false;
	    	callback();
	    	
	    }
	    
	} else {

    	callback();
		
	}
}