// hulpfunctie om methode aan een functie-prototype te koppelen
Function.prototype.method = function(name, func) {
    
    // koppel functie aan eigenschap van prototype met opgegeven naam indien
    // het prototype nog geen eigenschap had met die naam
    if (!this.prototype.hasOwnProperty(name)) {
        Object.defineProperty(
            this.prototype,
            name,
            {
                value: func,
                enumerable: false
            }
        );
    }
    
    // geef functie terug zodat method calls aaneengeschakeld kunnen worden
    return this;

}

//new array method that extends the current array with the elements from
//another array
Array.method("extend", function(other) {
	
	for (var i = 0; i < other.length; ++i) {
		this.push(other[i]);
	}
	
});

// nieuwe arraymethode die de elementen van de array willekeurig door elkaar
// schudt (implementeert Fisher-Yates algoritme in place)
// zie: http://bost.ocks.org/mike/shuffle/compare.html
Array.method('shuffle', function() {

    var huidigeIndex = this.length,
        tijdelijkeWaarde,
        willekeurigeIndex;

    while (huidigeIndex) {

        // kies willekeurige index die kleiner is dan de huidige index
        willekeurigeIndex = Math.floor(Math.random() * huidigeIndex);

        // schuif huidige index één positie naar links op; hierdoor kan de
        // willekeurig gekozen index gelijk zijn aan huidige index, waardoor
        // element op de huidige index op zijn plaats blijft staan
        huidigeIndex -= 1;

        // verwissel elementen op huidige index en willekeurig gekozen index die
        // kleiner of gelijk is aan huidige index
        tijdelijkeWaarde = this[huidigeIndex];
        this[huidigeIndex] = this[willekeurigeIndex];
        this[willekeurigeIndex] = tijdelijkeWaarde;
        
    }

    // geef een referentie naar de array terug, zodat arraymethoden kunnen
    // aaneengeschakeld worden
    return this;
    
});

// Array.prototype.reduce polyfill (MDN recommendation)
// Production steps of ECMA-262, Edition 5, 15.4.4.21
// Reference: http://es5.github.io/#x15.4.4.21
Array.method('reduce', function(callback /*, initialValue*/) {
	
	'use strict';
	
	if (this === null) {
		throw new TypeError('Array.prototype.reduce called on null or undefined');
	}
	
	if (typeof callback !== 'function') {
		throw new TypeError(callback + ' is not a function');
	}
	
	var t = Object(this), len = t.length >>> 0, k = 0, value;
	
	if (arguments.length === 2) {
		value = arguments[1];
	} else {
		
		while (k < len && !(k in t)) {
			k++; 
		}
		
		if (k >= len) {
			throw new TypeError('Reduce of empty array with no initial value');
		}
		
		value = t[k++];
	}
	
	for (; k < len; k++) {
		if (k in t) {
			value = callback(value, t[k], k, t);
		}
	}
	
	return value;
	
});

// nieuwe stringmethode die in de string alle plaatshouders van de vorm {naam}
// vervangt door de waarde die eraan gekoppeld wordt in het object dat als
// parameter wordt doorgegeven aan de methode; indien dit object geen eigenschap
// heeft met de opgegeven naam, dan wordt de plaatshouder niet vervangen
String.method('format', function (dict) {
    
    return this.replace(/{([^{}]*)}/g, function (match, naam) {
        var waarde = dict[naam];
        return waarde !== undefined ? waarde.toString() : match;
    });
    
});

// new string method that returns a Boolean value that indicates whether or not
// the string only contains letters of the alphabet
String.method("isAlpha", function() {
	
	return this.search(/^[a-z]*$/i) !== -1;
	
});

// new string method that returns a Boolean value that indicates whether or not
//t he string only contains digits and letters of the alphabet
String.method("isAlnum", function() {
	
	return this.search(/^[0-9a-z]*$/i) !== -1;
	
});

// new string method that returns a Boolean value that indicates whether or not
// the string contains lowercase letters only
String.method("isLower", function() {
	
	return this.search(/^[a-z]*$/) !== -1;
	
});

// new string method that returns a Boolean value that indicates whether or not
// the string contains lowercase letters only
String.method("isUpper", function() {
	
	return this.search(/^[A-Z]*$/) !== -1;
	
});